﻿#include "TRobotSerialIOWindows.hpp"
#include "TRobotSensorData.hpp"
#include "TRobotUtil.hpp"

#include <windows.h>

#include <iostream>
#include <algorithm>
#include <string>

#include <cstdio>
#include <cstdlib>
#include <cstring>


namespace{
  //////////////////////////////////////////////////////////////////////////////
  /*!
   *  @brief 現在の GetLastError() の値に対応するOS定義のメッセージを出力する
   *
   *  @copybrief
   *
   *  @param customMsg [in] 付加的に表示するメッセージ
   */
  void printLastError(const char* customMsg)
  {
    using namespace std;
  
    LPTSTR msgBuf;
    DWORD MIN_NUM_TCHARS = 0;

    FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
                  FORMAT_MESSAGE_FROM_SYSTEM |
                  FORMAT_MESSAGE_IGNORE_INSERTS,
                  NULL, // FROM_SYSTEM なので lpSource = NULL
                  GetLastError(),
                  //LANG_USER_DEFAULT,
                  MAKELANGID(LANG_ENGLISH, SUBLANG_ENGLISH_US), // 多言語対応は考慮しない
                  (LPTSTR)&msgBuf,
                  MIN_NUM_TCHARS,
                  NULL);

  
    cout << customMsg << ": " << msgBuf << endl;

    // FormatMessage() に FORMAT_MESSAGE_ALLOCATE_BUFFER を指定しているので
    // 以下の解放が必要
    LocalFree(msgBuf); 
  }

  //////////////////////////////////////////////////////////////////////////////
  /*!
   *  @brief スレッドが止まるまで待つ
   *
   *  @copybrief
   *
   *  @param thrHandle 対象のスレッドへのハンドル
   *  @param thrName   対象のスレッドの名称(メッセージ表示用) 
   */
  void joinThread(HANDLE thrHandle, const char *thrName)
  {
    using namespace std;
    DWORD retCode;
  
    cout << "Waiting for " << thrName << " thread to stop..." << endl;
    retCode = WaitForSingleObject(thrHandle, INFINITE);

    switch(retCode){
    case WAIT_OBJECT_0:
      cout << thrName << " thread successfully stopped." << endl;
      break;
    
    case WAIT_FAILED:
      cout << "Error detected on stopping " << thrName << " thread." << endl;
      break;
    }  
  
  }

  //////////////////////////////////////////////////////////////////////////////
  void closeHandleErrCheck(HANDLE handle)
  {
    BOOL res = FALSE;

    res = CloseHandle(handle);
    if(res == FALSE){
      printLastError("On closing a handle");
    }
  }  
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  コンストラクタ
 */
TRobot::SerialIOWindows::SerialIOWindows() :
  m_commPort(INVALID_HANDLE_VALUE),
  m_recvThrHandle(INVALID_HANDLE_VALUE),
  m_stopRequested(false)
{
  InitializeCriticalSection(&m_stopReqLock);
  InitializeCriticalSection(&m_senDataQLock);
  InitializeCriticalSection(&m_rotVelCmdQLock);
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  デストラクタ
 */
TRobot::SerialIOWindows::~SerialIOWindows()
{
  DeleteCriticalSection(&m_stopReqLock);
  DeleteCriticalSection(&m_senDataQLock);
  DeleteCriticalSection(&m_rotVelCmdQLock);
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 移動ロボットの基板との通信のために構造体DCBの設定を行う
 *
 *  @copybrief
 *
 *  DCB: Device Control Block
 *
 *  @param dcb [out] 設定対象の構造体DCB
 */
static void setupDCB(DCB& dcb)
{
  dcb.BaudRate     = CBR_115200;
  dcb.ByteSize     = 8;
  dcb.fParity      = FALSE; // パリティチェックなし
  dcb.StopBits     = ONESTOPBIT;  
  dcb.fOutxCtsFlow = TRUE;  // CTS/RTS フロー制御無効            
  dcb.fRtsControl  = RTS_CONTROL_DISABLE;

  dcb.fOutxDsrFlow = FALSE;              
  dcb.fDtrControl  = DTR_CONTROL_DISABLE;
  dcb.fDsrSensitivity		= FALSE;
  dcb.fTXContinueOnXoff	= FALSE;
  dcb.fOutX				  = FALSE;
  dcb.fInX				  = FALSE;
  dcb.fErrorChar	  = FALSE;
  dcb.fNull				  = FALSE;
  dcb.fAbortOnError	= FALSE;
  dcb.fDummy2				= 0;
  dcb.wReserved			= 0;
  dcb.XonLim				= 0;
  dcb.XoffLim				= 0;
  dcb.XonChar				= 0;
  dcb.XoffChar			= 0;
  dcb.ErrorChar			= 0;
  dcb.EofChar				= 0;
  dcb.EvtChar				= 0;
  dcb.wReserved1		= 0;  
}



////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 通信を開始する
 *
 *  @copybrief
 *
 *  @param port [in] 通信ポートを表す文字列
 *                   (Windows では COM10 など) 
 */
TRobot::RetCode TRobot::SerialIOWindows::start(const char *portName)
{
  using namespace std;
  const DWORD  NO_SHARE    = 0;
  const HANDLE NO_TEMPLATE = NULL;
  const SIZE_T USE_DEFAULT_STACK_SIZE = 0;
  const DWORD DEFAULT_CREATION_FLAGS = 0;
  const LPSECURITY_ATTRIBUTES DEFAULT_SECURITY_ATTRIB = NULL;
  DCB dcb;
  COMMTIMEOUTS timeouts;
  BOOL succeeded;

  // COM10 以上の番号を指定するために必要な接頭辞
  // (UNC: Universal Naming Convention)
  const string UNC_PREFIX = "\\\\.\\"; 
  string prefixPortName = UNC_PREFIX + portName;

  cout << "Opening " << prefixPortName << " for communication..." << endl;
  
  m_commPort = CreateFile(prefixPortName.c_str(),
                          GENERIC_READ | GENERIC_WRITE,
                          NO_SHARE,
                          DEFAULT_SECURITY_ATTRIB,
                          OPEN_EXISTING,
                          0, //FILE_ATTRIBUTE_NORMAL,
                          NO_TEMPLATE);
  
  if(m_commPort == INVALID_HANDLE_VALUE){
    printLastError("SerialIOWindows::start()");
    return TRobot::ERROR_OPEN_PORT;
  }
  cout << "Opened " << portName << endl;

  memset(&dcb, 0, sizeof(DCB));
  
  // 明示的に指定しない項目についてデフォルトの設定を引き継ぐため
  succeeded = GetCommState(m_commPort, &dcb);
  if(!succeeded){
    printLastError("SerialIOWindows::start()");
    return TRobot::ERROR_OPEN_PORT;
  }
  
  setupDCB(dcb);
  
  succeeded = SetCommState(m_commPort, &dcb);
  if(!succeeded){
    printLastError("SerialIOWindows::start()");
    return TRobot::ERROR_OPEN_PORT;
  }  


  // Time-out を無効にする
  timeouts.ReadIntervalTimeout         = 0; 
  timeouts.ReadTotalTimeoutMultiplier  = 0;
  timeouts.ReadTotalTimeoutConstant    = 0;
  timeouts.WriteTotalTimeoutMultiplier = 0;
  timeouts.WriteTotalTimeoutConstant   = 0;
  
  succeeded = SetCommTimeouts(m_commPort, &timeouts);
  if(!succeeded){
    printLastError("SerialIOWindows::start()");
    return TRobot::ERROR_OPEN_PORT;
  }
  cout << "Set time-out" << endl;

  

  m_recvThrHandle = CreateThread(DEFAULT_SECURITY_ATTRIB,
                                 USE_DEFAULT_STACK_SIZE,   
                                 recvThreadFuncWrapper,
                                 this,
                                 DEFAULT_CREATION_FLAGS,
                                 NULL); // thread id 不要

  if(m_recvThrHandle == NULL){
    printLastError("SerialIOWindows::start()");
    return TRobot::ERROR_CREATE_RECV_THREAD;
  }
  
  cout << "Created a receiver thread" << endl;

  
  m_sendThrHandle = CreateThread(DEFAULT_SECURITY_ATTRIB,
                                 USE_DEFAULT_STACK_SIZE,   
                                 sendThreadFuncWrapper,
                                 this,
                                 DEFAULT_CREATION_FLAGS,
                                 NULL); // thread id 不要
  
  if(m_sendThrHandle == NULL){
    printLastError("SerialIOWindows::start()");
    return TRobot::ERROR_CREATE_SEND_THREAD;
  }

  cout << "Created a sender thread" << endl;

  
  m_isRunning = true;
  
  return TRobot::SUCCESS;
}





////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 通信を終了する
 *
 *  @copybrief
 */
TRobot::RetCode TRobot::SerialIOWindows::stop()
{
  requestStop();

  joinThread(m_recvThrHandle, "receiver");
  joinThread(m_sendThrHandle, "sender");

  closeHandleErrCheck(m_recvThrHandle);
  closeHandleErrCheck(m_sendThrHandle);
  closeHandleErrCheck(m_commPort);

  m_stopRequested = false;
  m_isRunning = false;

  // 次にstart()が呼ばれるときに，残っているデータのせいで遅延が生じないようにする
  clearSensorDataQueue();
  clearCmdPairQueue();
  
  return TRobot::SUCCESS;
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 通信スレッドの停止要求の有無を確認する
 *
 *  @copybrief
 */
bool TRobot::SerialIOWindows::stopRequested()
{
  bool req = false;
  EnterCriticalSection(&m_stopReqLock);
  req = m_stopRequested;
  LeaveCriticalSection(&m_stopReqLock);

  return req;
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 通信スレッドの停止を要求する
 *
 *  @copybrief
 */
void TRobot::SerialIOWindows::requestStop()
{
  EnterCriticalSection(&m_stopReqLock);
  m_stopRequested = true;
  LeaveCriticalSection(&m_stopReqLock);
}



////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 指定された分のデータを全て受信する
 *
 *  @copybrief
 *
 *  @param port  [in] 通信ポートへのハンドル
 *  @param buf   [out] 受信データを格納するバッファ
 *  @param numBytes [in] 受信するデータのサイズ [bytes]
 */
TRobot::RetCode TRobot::SerialIOWindows::receiveExactNumBytes
(HANDLE port, unsigned char *buf, size_t numBytes)
{
  using namespace std;
  DWORD nrecvTotal = 0;
  DWORD nrecv = 0;
  LPOVERLAPPED NO_OVERLAP = NULL;
  BOOL succeeded;

  while(nrecvTotal < numBytes){
    succeeded = ReadFile(port,
                         buf + nrecvTotal,
                         numBytes - nrecvTotal,
                         &nrecv,
                         NO_OVERLAP);
    if(succeeded){
      nrecvTotal += nrecv;
      // cout << "Received " << nrecv << " [bytes]" << endl;
    }else{
      printLastError("SerialIOWindows::receiveExactNumBytes()");
    }
  }

  return TRobot::SUCCESS;
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 指定された分のデータを全て送信する
 *
 *  @copybrief
 *
 *  @param port     [in] 通信ポートのハンドル
 *  @param buf      [in] 送信データを格納するバッファ
 *  @param numBytes [in] 送信するデータのサイズ [bytes] 
 */
TRobot::RetCode TRobot::SerialIOWindows::sendExactNumBytes
(HANDLE port, unsigned char *buf, size_t numBytes)
{
  DWORD nsent = 0;
  DWORD nsentTotal = 0;
  LPOVERLAPPED NO_OVERLAP = NULL;
  BOOL succeeded = false;  

  while(nsentTotal < numBytes){
    succeeded = WriteFile(port,
                          buf + nsentTotal,
                          numBytes - nsentTotal,
                          &nsent,
                          NO_OVERLAP);
    if(succeeded){
      nsentTotal += nsent;
    }else{
      printLastError("SerialIOWindows::sendExactNumBytes()");
    }
  }  

#ifndef NDEBUG
  using namespace std;
  cerr << "sent: "; 
  for(size_t i = 0; i < nsent; i++){
    fprintf(stderr, "%02x ", buf[i]);
  }
  cerr << endl;
#endif

  return TRobot::SUCCESS;
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 受信スレッドのルーチンを扱うためのラッパ関数
 *
 *  @copybrief
 *
 *  非staticメンバ関数のコールバックは直接行えないので間を取り持つ
 *  static メンバ関数を使う。 
 
 *
 *  @pre SerialIOWindows のインスタンスが1個のみ存在する
 *       (現状，この関数は複数のインスタンスがアクセスすると処理が破綻する) 
 */
DWORD WINAPI TRobot::SerialIOWindows::recvThreadFuncWrapper(LPVOID params)
{
  const LPVOID DUMMY_PARAMS = NULL;
  SerialIOWindows *self;
  DWORD retVal = 0;
  
  if(params != NULL){
    self = (SerialIOWindows*) params;
    retVal = self->recvThreadFunc(DUMMY_PARAMS);
  }
  
  return retVal;
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *
 *  次のヘッダが見つかるまで読み込む
 *
 *  ポーリングでなく一方的なストリーミングのため，
 *  読み込んだバッファの先頭がデータのヘッダとは限らない
 *
 *  @param buf    [out]  読み込み用のバッファ
 *  @param bufLen [in]   buf の長さ [bytes]
 */
void TRobot::SerialIOWindows::seekToNextHeader(unsigned char *buf,
                                               size_t bufLen)
{
  using namespace std;
  const unsigned char *headPos = NULL;
  unsigned char *bufLast = &buf[bufLen];
  size_t remDataLen = 0;
  const DWORD WAIT_DURATION = 8; // [msec]
  
  while(true){ // ヘッダが見つかるまで
    receiveExactNumBytes(m_commPort, buf, BASE_SENSOR_DATA_LEN);
    
    headPos = searchForValidHeader(buf, bufLast);
    if(headPos != bufLast){
      break;
    }

    Sleep(WAIT_DURATION);
  }

  remDataLen = headPos - buf; // ヘッダまでの残りのデータ長
  // cout << "remDataLen == " << remDataLen << endl;

  // remDataLen == 0 であれば buf の先頭にヘッダがある
  if(remDataLen != 0){
    // 次のヘッダの直前まで読み込んで捨てる
    receiveExactNumBytes(m_commPort, buf, remDataLen);
    
    receiveExactNumBytes(m_commPort, buf, BASE_SENSOR_DATA_LEN);
    Sleep(WAIT_DURATION);
  }
}
                                             


////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 受信スレッドのメインルーチン 
 *
 *  @copybrief
 */
DWORD TRobot::SerialIOWindows::recvThreadFunc(LPVOID params)
{
  using namespace TRobot;
  using namespace std;
  unsigned char buf[BASE_SENSOR_DATA_LEN];
  BaseSensorData sensorData;
  const DWORD WAIT_DURATION = 8; // [msec]
  const DWORD THREAD_EXIT_SUCCESS = 0;

  DBGOUT(<< "recv thread starting..." << endl);
  while(!stopRequested()){
    seekToNextHeader(buf, BASE_SENSOR_DATA_LEN);
    decodeBaseSensorData(buf, sensorData);
    pushSensorData(sensorData);

    Sleep(WAIT_DURATION);
  }
  
  return THREAD_EXIT_SUCCESS;
}




////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 送信スレッド関数
 *
 *  @copybrief
 */
void TRobot::SerialIOWindows::sendThreadFunc()
{
  using namespace std;
  static unsigned char cmd[TRobot::ROT_VEL_CMD_LEN];
  RetCode retCode;
  const DWORD WAIT_DURATION = 8; // [msec]
  const DWORD THREAD_EXIT_SUCCESS = 0;
  RotVelCmdPair cmdPair;
  
  while(!stopRequested()){
    if(cmdPairQueueEmpty()){
      continue;
    }
    popCmdPair(cmdPair);
    
    retCode = buildCmdSeqSetRotVel(cmdPair.left(), cmdPair.right(), cmd);
    if(retCode != TRobot::SUCCESS){
      cerr << "Failed to build a command sequence" << endl;
      break;
    }

    sendExactNumBytes(m_commPort, cmd, TRobot::ROT_VEL_CMD_LEN);

    Sleep(WAIT_DURATION);
  }
  
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief インスタンスメンバの sendThreadFunc() を呼び出すためのラッパ
 *
 *  @copybrief
 */
DWORD WINAPI TRobot::SerialIOWindows::sendThreadFuncWrapper(LPVOID params)
{
  SerialIOWindows *self;
  
  if(params != NULL){
    self = (SerialIOWindows*) params;
    self->sendThreadFunc();
  }
  
  return 0;
}

////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 左右の車輪の回転速度の目標値を送信する
 *
 *  @copybrief
 *
 *  @param leftRotVel  左車輪の回転速度指令値
 *  @param rightRotVel 右車輪の回転速度指令値
 */
TRobot::RetCode TRobot::SerialIOWindows::sendWheelRotVel
(int leftRotVel, int rightRotVel)
{
  pushCmdPair(RotVelCmdPair(leftRotVel, rightRotVel));
  
  return TRobot::SUCCESS;  
}

////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief センサデータを受信する
 *
 *  @copybrief
 *
 *  @param sensorData [out] センサデータ
 */
TRobot::RetCode TRobot::SerialIOWindows::receiveSensorData
(TRobot::BaseSensorData& sensorData)
{
  static const TRobot::BaseSensorData NIL_SENSOR_DATA;

  if(sensorDataQueueEmpty()){
    return TRobot::ERROR_QUEUE_EMPTY;
    // sensorData = NIL_SENSOR_DATA;
  }
  // empty() -> pop() の間で排他処理をしないための前提:
  //  この関数を実行中のスレッド以外は popSensorData()を呼び出さない。
  //  他のスレッドがpush()を行うのは可
  //   (push()してもキューは空にならないので popSensorData()で失敗しない) 。
  
  popSensorData(sensorData);
  
  return TRobot::SUCCESS;  
}



////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief センサデータをキューに入れる
 *
 *  @copybrief
 *
 *  スレッドセーフ
 */
void TRobot::SerialIOWindows::pushSensorData(const BaseSensorData& sensorData)
{
  EnterCriticalSection(&m_senDataQLock);
  m_sensorDataQueue.push_back(sensorData);
  LeaveCriticalSection(&m_senDataQLock);
}

////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief センサデータをキューから取り出す
 *
 *  @copybrief
 *
 *  スレッドセーフ
 */
void TRobot::SerialIOWindows::popSensorData(BaseSensorData& sensorData)
{
  EnterCriticalSection(&m_senDataQLock);
  sensorData = m_sensorDataQueue.front();
  m_sensorDataQueue.pop_front();    
  LeaveCriticalSection(&m_senDataQLock);
}

////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief センサデータのキューが空か否かを確認する
 *
 *  @copybrief
 * 
 *  スレッドセーフ
 *
 *  @return 空であれば true，空でなければ false
 */
bool TRobot::SerialIOWindows::sensorDataQueueEmpty()
{
  bool empty = false;
  EnterCriticalSection(&m_senDataQLock);
  empty = m_sensorDataQueue.empty();
  LeaveCriticalSection(&m_senDataQLock);
  
  return empty;
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief センサデータのキューを空にする
 *
 *  @copybrief
 * 
 *  スレッドセーフ
 */
void TRobot::SerialIOWindows::clearSensorDataQueue()
{
  EnterCriticalSection(&m_senDataQLock);
  m_sensorDataQueue.clear();
  LeaveCriticalSection(&m_senDataQLock);
}
  

////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 指令値をキューに入れる
 *
 *  @copybrief
 *
 *  スレッドセーフ
 */
void TRobot::SerialIOWindows::pushCmdPair(const RotVelCmdPair& cmdPair)
{
  EnterCriticalSection(&m_rotVelCmdQLock);
  m_rotVelCmdQueue.push_back(cmdPair);
  LeaveCriticalSection(&m_rotVelCmdQLock);
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 指令値をキューから取り出す
 *
 *  @copybrief
 *
 *  スレッドセーフ
 */
void TRobot::SerialIOWindows::popCmdPair(RotVelCmdPair& cmdPair)
{
  EnterCriticalSection(&m_rotVelCmdQLock);
  cmdPair = m_rotVelCmdQueue.front();
  m_rotVelCmdQueue.pop_front();    
  LeaveCriticalSection(&m_rotVelCmdQLock);
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 指令値のキューが空か否かを確認する
 *
 *  @copybrief
 * 
 *  スレッドセーフ
 *
 *  @return 空であれば true，空でなければ false
 */
bool TRobot::SerialIOWindows::cmdPairQueueEmpty()
{
  bool empty = false;
  EnterCriticalSection(&m_rotVelCmdQLock);
  empty = m_rotVelCmdQueue.empty();
  LeaveCriticalSection(&m_rotVelCmdQLock);
  
  return empty;  
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 指令値のキューを空にする
 *
 *  @copybrief
 * 
 *  スレッドセーフ
 */
void TRobot::SerialIOWindows::clearCmdPairQueue()
{
  EnterCriticalSection(&m_rotVelCmdQLock);
  m_rotVelCmdQueue.clear();
  LeaveCriticalSection(&m_rotVelCmdQLock);  
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *
 *  @return 通信中であれば true，そうでなければ false
 */
bool TRobot::SerialIOWindows::isRunning()
{
  return m_isRunning;
}
